#include "stdafx.h"
#include "DriverInstaller.h"
#include "ModelsTargetOSVer.h"
#include "HardwareID.h"

typedef std::map<HardwareID, HardwareID> HardwareIDMap;
typedef std::map<std::wstring, ModelsTargetOSVer> ModelsCollection;

namespace fs {
	using namespace boost::filesystem;
};

bool matchesTargetMoreClosely(
	unsigned int new_architecture_flag, uint32_t new_version,
	unsigned int current_architecture_flag, uint32_t current_version,
	std::wstring const &architecture_str, unsigned int version);
ModelsCollection getModelsSections(
	MyLib::InfFile &inf_file,
	int target_architecture,
	unsigned int version);


DriverInstallationResult installDriver(
	HWND window_handle,
	fs::wpath const &inf_file_path,
	std::wstring const &hwid_str,
	int architecture_id,
	uint32_t os_version)
{
	HardwareID const hwid(hwid_str);

	int hwid_match_level = 0;
	std::wstring ddinstall_section;
	//std::wstring hw_id_str;
	MyLib::InfFile inf_file(inf_file_path.c_str()/*.wstring().c_str()*/);
	ModelsCollection models_collection = getModelsSections(inf_file, architecture_id, os_version);
	for (ModelsCollection::const_iterator it = models_collection.begin();
		it != models_collection.end();
		++it) {
			std::wstring models_section = it->first + it->second.getVerString();
			//wprintf(L"models_section : %s\n", models_section.c_str());
			MyLib::InfContextIterator it1 = inf_file.findFirstLine(models_section.c_str());
			MyLib::InfContextIterator const it1_end;
			for (; it1 != it1_end; ++it1) {
				unsigned int const field_count = it1->getFieldCount();
				if (field_count >= 2) {
					std::wstring inf_hwid_str = it1->getStringField(2);
					if (inf_hwid_str.empty() && field_count >= 3) {
						inf_hwid_str = it1->getStringField(3);
					}
					HardwareID inf_hwid(inf_hwid_str);
					int hwid_match_level_temp = hwid.match(inf_hwid);
					if (hwid_match_level_temp > hwid_match_level) {
						ddinstall_section = it1->getStringField(1);
						//hw_id_str = hw_id_str0;
						//wprintf(L"ddinstall_section : %s\n", ddinstall_section.c_str());
						hwid_match_level = hwid_match_level_temp;
					}
				}
			}
	}
	if (ddinstall_section.empty()) {
		return INVALID_INF_FILE;
	}

	BOOL ret;
	DWORD section_name_size;
	ret = ::SetupDiGetActualSectionToInstall(inf_file, ddinstall_section.c_str(), NULL, 0, &section_name_size, NULL);
	if (!ret) {
		BOOST_THROW_EXCEPTION(MyLib::SetupAPIException());
	}
	boost::scoped_array<wchar_t> section_name_buf(new wchar_t[section_name_size]);
	::SetupDiGetActualSectionToInstall(inf_file, ddinstall_section.c_str(), section_name_buf.get(), section_name_size, NULL, NULL);
	std::wstring const section_name = section_name_buf.get();

	ret = SetupCopyOEMInf(inf_file_path.c_str(), NULL, SPOST_PATH, 0, NULL, 0, NULL, NULL);
	if (!ret) {
		//throw MY_SETUP_API_EXCEPTION_LAST_ERROR;
		BOOST_THROW_EXCEPTION(MyLib::SetupAPIException());
	}

	HSPFILEQ q = ::SetupOpenFileQueue();
	void *context = ::SetupInitDefaultQueueCallback(window_handle);
	ret = ::SetupInstallFilesFromInfSection(inf_file, NULL, q, section_name.c_str(), NULL, SP_COPY_NEWER);
	if (!ret) {
		::SetupTermDefaultQueueCallback(context);
		::SetupCloseFileQueue(q);
		//throw MY_SETUP_API_EXCEPTION_LAST_ERROR;
		BOOST_THROW_EXCEPTION(MyLib::SetupAPIException());
	}
	::SetupCommitFileQueue(window_handle, q, SetupDefaultQueueCallback, context);
	::SetupTermDefaultQueueCallback(context);
	::SetupCloseFileQueue(q);

	//ret = SetupInstallFromInfSection(NULL, hinf, actual_inf_section, SPINST_ALL, HKEY_LOCAL_MACHINE, NULL, SP_COPY_FORCE_IN_USE, SetupDefaultQueueCallback, context, NULL, NULL);

	//wcscat(actual_inf_section, L".Services");
	std::wstring const services_section_name = section_name + L".Services";
	ret = ::SetupInstallServicesFromInfSection(inf_file, services_section_name.c_str(), 0);
	if (!ret) {
		//throw MY_SETUP_API_EXCEPTION_LAST_ERROR;
		BOOST_THROW_EXCEPTION(MyLib::SetupAPIException());
	}

	////////////////////////////////////////////////////////////////////////////
	// CriticalDeviceDatabase
	////////////////////////////////////////////////////////////////////////////

	//HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\CriticalDeviceDatabase\pci#ven_8086&dev_27c5&cc_0106
	static std::wstring const REGSTR_PATH_CRITICALDEVICEDATABASE_BS = L"System\\CurrentControlSet\\Control\\CriticalDeviceDatabase\\";
	//static size_t const REGSTR_PATH_CRITICALDEVICEDATABASE_BS_LEN = wcslen(REGSTR_PATH_CRITICALDEVICEDATABASE_BS);

	std::wstring const hwid_str2 = hwid.getString();
	size_t const hwid_str2_len = hwid_str2.length();
	boost::scoped_array<wchar_t> hwid_str2_buf(new wchar_t[hwid_str2_len + 1]);
	memcpy(hwid_str2_buf.get(), hwid_str2.c_str(), (hwid_str2_len + 1)*sizeof(wchar_t));
	// ɂ
	::CharLower(hwid_str2_buf.get());
	// \  # ɕύX
	wchar_t * const bs_ptr = ::wcschr(hwid_str2_buf.get(), L'\\');
	if (bs_ptr != NULL) {
		*bs_ptr = L'#';
	}
	std::wstring const reg_key_name = REGSTR_PATH_CRITICALDEVICEDATABASE_BS + hwid_str2_buf.get();

	//boost::scoped_array<wchar_t> reg_key_buf(new wchar_t[REGSTR_PATH_CRITICALDEVICEDATABASE_BS_LEN + hwid_str2.length() + 1]);
	//memcpy(reg_key_buf.get(), REGSTR_PATH_CRITICALDEVICEDATABASE_BS, (REGSTR_PATH_CRITICALDEVICEDATABASE_BS_LEN + 1)*sizeof(wchar_t));
	//// 10/24
	//wcscat(reg_key_buf.get(), ahci_id.getString().c_str());
	////wcscat(reg_key_buf.get(), hw_id_str.c_str());
	//::CharLower(reg_key_buf.get() + REGSTR_PATH_CRITICALDEVICEDATABASE_BS_LEN);
	//wchar_t * const bs_ptr = ::wcschr(reg_key_buf.get() + REGSTR_PATH_CRITICALDEVICEDATABASE_BS_LEN, L'\\');
	//if (bs_ptr != NULL) {
	//	*bs_ptr = L'#';
	//}

	MyLib::RegistryKey reg_key = MyLib::RegistryKey::createKey(HKEY_LOCAL_MACHINE, reg_key_name.c_str());
	//HKEY hk;
	//DWORD dw;
	//long l = ::RegCreateKeyEx(
	//	HKEY_LOCAL_MACHINE,
	//	reg_key_buf.get(),
	//	0,
	//	REG_NONE,
	//	REG_OPTION_NON_VOLATILE,
	//	KEY_READ | KEY_WRITE,
	//	NULL,
	//	&hk,
	//	&dw);
	//if (l != NO_ERROR) {
	//	throw MY_WIN32_EXCEPTION(l);
	//}

	std::wstring const class_guid = inf_file.getVersionInfo(L"ClassGuid");
	//l = ::RegSetValueEx(hk, L"ClassGUID", NULL, REG_SZ,
	//	reinterpret_cast<const BYTE*>(class_guid.c_str()),
	//	(class_guid.length() + 1)*sizeof(wchar_t));
	//if (l != NO_ERROR) {
	//	::RegCloseKey(hk);
	//	throw MY_WIN32_EXCEPTION(l);
	//}
	reg_key.setStringValue(L"ClassGUID", class_guid.c_str());

	MyLib::InfContextIterator it1 = inf_file.findFirstLine(services_section_name.c_str());
	std::wstring const service_name = it1->getStringField(1);
	//l = ::RegSetValueEx(hk, L"Service", NULL, REG_SZ,
	//	reinterpret_cast<const BYTE*>(service_name.c_str()),
	//	(service_name.length() + 1)*sizeof(wchar_t));
	//if (l != NO_ERROR) {
	//	::RegCloseKey(hk);
	//	throw MY_WIN32_EXCEPTION(l);
	//}
	reg_key.setStringValue(L"Service", service_name.c_str());

	return SUCCESS;
}

bool matchesTargetMoreClosely(
	unsigned int new_architecture_flag, uint32_t new_version,
	unsigned int current_architecture_flag, uint32_t current_version,
	std::wstring const &architecture_str, unsigned int version)
{
	unsigned int this_architecture_flag =
		ModelsTargetOSVer::getArchitectureFlagFromString(architecture_str);
	if ((new_architecture_flag & this_architecture_flag)) {
		if ((current_version == new_version && current_architecture_flag > new_architecture_flag) ||
			(current_version < new_version && new_version <= version)) {
				return true;
		}
	}
	return false;
}

ModelsCollection getModelsSections(MyLib::InfFile &inf_file, int target_architecture, unsigned int version)
{
	std::wstring architecture_str;
	switch (target_architecture) {
	case PROCESSOR_ARCHITECTURE_INTEL:
		architecture_str = _T("X86");
		break;
	case PROCESSOR_ARCHITECTURE_AMD64:
		architecture_str = _T("AMD64");
		break;
	case PROCESSOR_ARCHITECTURE_IA64:
		architecture_str = _T("IA64");
		break;
	default:
		throw std::runtime_error("̃vZbTɂ͑ΉĂ܂");
	}

	ModelsCollection models_collection;
	const MyLib::InfContextIterator inf_end;
	const ModelsTargetOSVer models_no_decorated;
	const unsigned int no_decorated_architecture_flags =
		models_no_decorated.getArchitectureFlag();
	// ^[Qbg CPU Architecture tO (X86, AMD64, IA64)
	const unsigned int target_architecture_flag =
		ModelsTargetOSVer::getArchitectureFlagFromString(architecture_str);

	MyLib::InfContextIterator inf_it = inf_file.findFirstLine(_T("Manufacturer"));
	for (; inf_it != inf_end; ++inf_it) {
		unsigned int field_count = inf_it->getFieldCount();
		// o[WȂǂtĂȂModels Section擾
		std::wstring models_base = inf_it->getStringField(field_count > 0 ? 1 : 0);
		ModelsCollection::iterator it = models_collection.find(models_base);
		if (it == models_collection.end()
			&& (no_decorated_architecture_flags & target_architecture_flag)) {
				it = models_collection.insert(
					std::make_pair(models_base, models_no_decorated)).first;
		}
		// XP(Ver.5.1)ȍ~TargetOSVersionw肳ĂƂ
		if (field_count >= 2 && version >= 0x00050001) {
			// TargetOSVersion(tB[hȍ~)̉
			for (unsigned int i = 2; i <= field_count; ++i) {
				ModelsTargetOSVer new_os_ver(inf_it->getStringField(i));
				// ɋ߂̂
				if (it == models_collection.end()) {
					if (matchesTargetMoreClosely(
						new_os_ver.getArchitectureFlag(),
						new_os_ver.getVersion(),
						models_no_decorated.getArchitectureFlag(),
						models_no_decorated.getVersion(),
						architecture_str,
						version)) {
							it = models_collection.insert(
								std::make_pair(models_base, new_os_ver)).first;
					}
				} else {
					if (matchesTargetMoreClosely(
						new_os_ver.getArchitectureFlag(),
						new_os_ver.getVersion(),
						it->second.getArchitectureFlag(),
						it->second.getVersion(),
						architecture_str,
						version)) {
							it->second = new_os_ver;
					}
				}
			}
		}
	}
	return models_collection;
}


//std::wstring createLogFileName()
//{
//	std::wstringstream ss;
//	ss.imbue(std::locale(""));
//	tm *t = localtime(time(NULL));
//	const std::time_put<TCHAR>& tp =
//		std::use_facet<std::time_put<TCHAR> >(ss.getloc());
//	LPCTSTR fmt = _T("%x");
//	tp.put(ss, ss, ss.fill(),
//		t, fmt, fmt + 2);
//}
