// AHCITool.cpp : R\[ AvP[ṼGg |Cg`܂B
//

#include "stdafx.h"
#include "DriverInstaller.h"

namespace fs {
	using namespace boost::filesystem;
};
namespace po {
	using namespace boost::program_options;
};

static wchar_t const *START_ENTRY_NAME = L"Start";
static uint32_t const OS_VERSION_VISTA = 0x00060000;
static wchar_t const *REGSTR_PATH_SERVICES_BS = L"System\\CurrentControlSet\\Services\\";

static void displayUsage();
bool isWOW64();
int getArchitectureID();
uint32_t getOSVersion();
fs::wpath const & getWinDirPath();

static void installInfFile(fs::wpath const &inf_file_path, std::wstring const &hwid, bool is_batch_mode);
static void setServiceBootStart(std::wstring const &driver_name, bool is_batch_mode);
static void enable_pciide(bool is_batch_mode);
static bool isDriverEnabled(std::wstring const &driver_name);

int wmain(int argc, wchar_t *argv[])
{
	_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

	if (!::IsUserAnAdmin()) {
		std::cout << "This program requires Administrator privileges." << std::endl;
		return EXIT_FAILURE;
	}

	if (isWOW64()) {
		std::cout << "This program cannot be run in WOW64 mode." << std::endl;
		return EXIT_FAILURE;
	}

	// R}hC
	//  ahcitool -i t@C -h n[hEFAID [-b]
	//  ahcitool -e {msahci|iastorv|pciide} [-b]
	//  ahcitool -c {msahci|iastorv|pciide}
	try {
		po::options_description opt("Options");
		opt.add_options()
			("install,i", po::wvalue<std::wstring>(), "install item")
			("enable,e", po::wvalue<std::wstring>(), "enable")
			("check,c", po::wvalue<std::wstring>(), "check enable")
			("batch,b", "batch")
			("hwid,h", po::wvalue<std::wstring>(), "hardware ids");

		po::variables_map vm;
		po::store(po::parse_command_line(argc, argv, opt), vm);
		po::notify(vm);

		bool const exists_i = vm.count("install") > 0;
		bool const exists_e = vm.count("enable") > 0;
		bool const exists_c = vm.count("check") > 0;
		bool const exists_b = vm.count("batch") > 0;
		bool const exists_h = vm.count("hwid") > 0;
		if (exists_i) {
			if (!exists_e && !exists_c && exists_h) {
				std::wstring str = vm["install"].as<std::wstring>();
				fs::wpath inf_file_path(str);
				std::wstring hwid = vm["hwid"].as<std::wstring>();
				installInfFile(inf_file_path, hwid, exists_b);
			} else {
				displayUsage();
			}
		} else if (exists_e) {
			if (!exists_h && !exists_c) {
				std::wstring str = vm["enable"].as<std::wstring>();
				if (_wcsicmp(str.c_str(), L"msahci") == 0) {
					setServiceBootStart(L"msahci", exists_b);
				} else if (_wcsicmp(str.c_str(), L"iastorv") == 0) {
					setServiceBootStart(L"iaStorV", exists_b);
				} else if (_wcsicmp(str.c_str(), L"pciide") == 0) {
					enable_pciide(exists_b);
				} else {
					displayUsage();
				}
			} else {
				displayUsage();
			}
		} else if (exists_c) {
			if (!exists_b && !exists_h) {	// -c-b/-h̓ws
				std::wstring str = vm["check"].as<std::wstring>();
				if (_wcsicmp(str.c_str(), L"msahci") == 0) {
					std::cout << (isDriverEnabled(L"msahci") ? "msahci is enabled." : "msahci is disabled.") << std::endl;
				} else if (_wcsicmp(str.c_str(), L"iaStorV") == 0) {
					std::cout << (isDriverEnabled(L"iaStorV") ? "iaStorV is enabled." : "iaStorV is disabled.") << std::endl;
				} else if (_wcsicmp(str.c_str(), L"pciide") == 0) {
					std::cout << ((isDriverEnabled(L"pciide") && isDriverEnabled(L"atapi")) ? "pciide is enabled." : "pciide is disabled.") << std::endl;
				} else {
					displayUsage();
				}
			} else {
				displayUsage();
			}
		} else {
			displayUsage();
		}

	} catch (MyLib::Exception &e) {
		std::cout << "ERROR!" << std::endl;
		std::cout << "err_code:" << e.getErrorCode() << std::endl;
		std::cout << "err_msg:" << e.what() << std::endl;
		int const *pn = boost::get_error_info<MyLib::errno_info>(e);
		if (pn) {
			std::cout << *pn << std::endl;
		}
		std::wstring const *pm = boost::get_error_info<MyLib::errmsg_info>(e);
		if (pm) {
			std::wcout.imbue(std::locale(""));
			std::wcout << *pm << std::endl;
		}
		return EXIT_FAILURE;
	} catch (boost::exception_detail::clone_impl<boost::exception_detail::error_info_injector<boost::program_options::invalid_command_line_syntax> > & /*e*/) {
		displayUsage();
	} catch (boost::exception &e) {
		std::cout << "ERROR!" << std::endl;
		std::cout << boost::diagnostic_information(e) << std::endl;
		return EXIT_FAILURE;
	} catch (std::exception &e) {
		std::cout << "ERROR!" << std::endl;
		std::cout << e.what() << std::endl;
		return EXIT_FAILURE;
	} catch (...) {
		std::cout << "UNKNOWN ERROR!" << std::endl;
		return EXIT_FAILURE;
	}
	return 0;
}

////////////////////////////////////////////////////////////////////////////////

static void displayUsage()
{
	//  ahcitool -i inf_file -h hardware_id [-b]
	//  ahcitool -e {msahci|pciide|intelide} [-b]
	//  ahcitool -c {msahci|pciide|intelide}
	std::cout << "Usage:" << std::endl;
	std::cout << "ahcitool -i inf_file -h hardware_id [-b]" << std::endl;
	std::cout << "ahcitool -e {pciide|msahci|iaStorV} [-b]" << std::endl;
	std::cout << "ahcitool -c {pciide|msahci|iaStorV}\n" << std::endl;
	std::cout << "  -i inf_file     INF File" << std::endl;
	std::cout << "  -h hardware_id  Hardware ID " << std::endl;
	std::cout << "  -e              Enables standard driver" << std::endl;
	std::cout << "  -b              Batch mode(Suppresses prompting to confirm you want to ...)" << std::endl;
	std::cout << "  -c driver_name  Check if driver_name is enabled" << std::endl;
}

static void installInfFile(fs::wpath const &inf_file_path, std::wstring const &hwid, bool is_batch_mode)
{
	if (!fs::exists(inf_file_path)) {
		// FILE NOT FOUND
		std::wcout.imbue(std::locale(""));
		std::wcout << L"Not found:\n" << inf_file_path.wstring() << std::endl;
		return;
	}
	if (!is_batch_mode) {
		int ch;
		for (;;) {
			char c;
			std::wcout << L"install inf file ? (y/n): ";
			std::cin >> c;
			if (!std::cin.fail()) {
				ch = std::tolower(c, std::locale(""));
				if (ch == 'y') {
					break;
				}
				if (ch == 'n') {
					return;
				}
			}
			std::cin.clear();
			std::cin.ignore(1024, '\n');
		}
	}
	int const architecture_id = getArchitectureID();
	uint32_t const os_version = getOSVersion();
	enum DriverInstallationResult ret = installDriver(NULL, inf_file_path, hwid, architecture_id, os_version);
	switch (ret) {
		case SUCCESS:
			std::cout << "The inf file has been successfully installed." << std::endl;
			break;
		case INVALID_INF_FILE:
			std::cout << "The inf file is corrupted." << std::endl;
			break;
		default:
			break;
	};
}

static void setServiceBootStart(std::wstring const &driver_name, bool is_batch_mode)
{
	// OS̃o[W`FbN
	uint32_t const os_version = getOSVersion();
	if (os_version < OS_VERSION_VISTA) {
		std::cout << "This function requires Vista or lator." << std::endl;
		return;
	}
	if (!is_batch_mode) {
		int ch;
		for (;;) {
			char c;
			std::wcout << L"enable " << driver_name << L" ? (y/n): ";
			std::cin >> c;
			if (!std::cin.fail()) {
				ch = std::tolower(c, std::locale(""));
				if (ch == 'y') {
					break;
				}
				if (ch == 'n') {
					return;
				}
			}
			std::cin.clear();
			std::cin.ignore(1024, '\n');
		}
	}
	std::wstring const msahci_reg_path = REGSTR_PATH_SERVICES_BS + driver_name;
	try {
		MyLib::RegistryKey reg_key =
			MyLib::RegistryKey::createKey(HKEY_LOCAL_MACHINE, msahci_reg_path.c_str());
		reg_key.setDWORDValue(L"Start", SERVICE_BOOT_START);
	} catch (MyLib::Exception &e) {
		e << MyLib::errmsg_info(L"HKLM\\" + msahci_reg_path);
		throw;
	}
	std::wcout << driver_name << L" has been enabled." << std::endl;
}

static void enable_pciide(bool is_batch_mode)
{
	if (!is_batch_mode) {
		int ch;
		for (;;) {
			char c;
			std::cout << "enable pciide? (y/n): ";
			std::cin >> c;
			if (!std::cin.fail()) {
				ch = std::tolower(c, std::locale(""));
				if (ch == 'y') {
					break;
				}
				if (ch == 'n') {
					return;
				}
			}
			std::cin.clear();
			std::cin.ignore(1024, '\n');
		}
	}
	int const architecture_id = getArchitectureID();
	uint32_t const os_ver = getOSVersion();
	fs::wpath const inf_path = getWinDirPath()/L"inf\\mshdc.inf";
	enum DriverInstallationResult ret = installDriver(NULL, inf_path, L"PCI\\CC_0101", architecture_id, os_ver);
	if (ret == SUCCESS) {
		ret = installDriver(NULL, inf_path, L"Primary_IDE_Channel", architecture_id, os_ver);
		if (ret == SUCCESS) {
			ret = installDriver(NULL, inf_path, L"Secondary_IDE_Channel", architecture_id, os_ver);
		}
	}
	switch (ret) {
		case SUCCESS:
			std::cout << "pciide has been enabled." << std::endl;
			break;
		case INVALID_INF_FILE:
			std::cout << "mshdc.inf is corrupted." << std::endl;
			break;
		default:
			break;
	};
}

static bool isDriverEnabled(std::wstring const &driver_name)
{
	std::wstring reg_key_name = REGSTR_PATH_SERVICES_BS + driver_name;
	try {
		if (MyLib::RegistryKey::isKeyExists(HKEY_LOCAL_MACHINE, reg_key_name.c_str())) {
			MyLib::RegistryKey reg_key(HKEY_LOCAL_MACHINE, reg_key_name.c_str(), KEY_READ);
			// Start
			if (reg_key.getEntryDataType(START_ENTRY_NAME) == REG_DWORD) {
				DWORD const start_value = reg_key.getDWORDValue(START_ENTRY_NAME);
				return start_value == SERVICE_BOOT_START;
			}
			return false;
		}
	} catch (MyLib::Exception &e) {
		e << MyLib::errmsg_info(reg_key_name);
		throw;
	}
	return false;
}

bool isWOW64()
{
	typedef BOOL (WINAPI* LPFN_ISWOW64PROCESS)(HANDLE hProcess, BOOL *Wow64Process);
	static LPFN_ISWOW64PROCESS fnIsWow64Process =
		reinterpret_cast<LPFN_ISWOW64PROCESS>(::GetProcAddress(
		::GetModuleHandle(L"kernel32.dll"), "IsWow64Process"));
	if (fnIsWow64Process != NULL) {
		BOOL is_wow64_process;
		fnIsWow64Process(::GetCurrentProcess(), &is_wow64_process);
		return is_wow64_process != FALSE;
	}
	return false;
}

int getArchitectureID()
{
	typedef void (WINAPI* LPFN_GETNATIVESYSTEMINFO)(SYSTEM_INFO *lpSystemInfo);
	static LPFN_GETNATIVESYSTEMINFO fnGetNativeSystemInfo =
		reinterpret_cast<LPFN_GETNATIVESYSTEMINFO>(::GetProcAddress(
		::GetModuleHandle(L"kernel32.dll"), "GetNativeSystemInfo"));
	if (fnGetNativeSystemInfo != NULL) {
		SYSTEM_INFO info;
		fnGetNativeSystemInfo(&info);
		return info.wProcessorArchitecture;
	}
	return PROCESSOR_ARCHITECTURE_INTEL;
}

uint32_t getOSVersion()
{
	// OS̃o[W`FbN
	OSVERSIONINFO osverinfo;
	osverinfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	::GetVersionEx(&osverinfo);
	return static_cast<uint32_t>(MAKELONG(osverinfo.dwMinorVersion, osverinfo.dwMajorVersion));
}

fs::wpath const & getWinDirPath()
{
	static fs::wpath *dir_path = NULL;
	if (dir_path == NULL) {
		unsigned int n = ::GetWindowsDirectory(NULL, 0);
		boost::scoped_array<wchar_t> buf(new wchar_t[n]);
		::GetWindowsDirectory(buf.get(), n);
		dir_path = new fs::wpath(buf.get());
	}
	return *dir_path;
}
