#include "StdAfx.h"
#include "WMIAccess.h"

/////////////////////////////////////////////////////////////////////////////
// WMIObject
/////////////////////////////////////////////////////////////////////////////
WMIObject::WMIObject(IWbemClassObject *cls_obj_ptr)
: cls_obj_ptr_(cls_obj_ptr)
{
}

HRESULT WMIObject::get(const wchar_t *name, VARIANT *val_ptr)
{
	return cls_obj_ptr_->Get(name, 0, val_ptr, 0, 0);
}

/////////////////////////////////////////////////////////////////////////////
// WMIObjectSet
/////////////////////////////////////////////////////////////////////////////
WMIObjectSet::WMIObjectSet()
: enumerator_ptr_(NULL),
  cls_obj_ptr_(NULL)
{
}

WMIObjectSet::WMIObjectSet(IEnumWbemClassObject *enum_ptr)
: enumerator_ptr_(enum_ptr),
  cls_obj_ptr_(NULL)
{
	next();
}

WMIObjectSet::~WMIObjectSet()
{
	if (enumerator_ptr_ != NULL) {
		enumerator_ptr_->Release();
	}
}

WMIObject WMIObjectSet::operator*()
{
	return cls_obj_ptr_;
}

//IWbemClassObject * WMIObjectSet::operator->()
//{
//	return cls_obj_ptr_;
//}

WMIObjectSet & WMIObjectSet::operator++()
{
	if (cls_obj_ptr_ != NULL) {
		cls_obj_ptr_->Release();
		cls_obj_ptr_ = NULL;
	}
	next();
	return *this;
}

bool WMIObjectSet::operator!=(WMIObjectSet const &crh) const
{
	return this->cls_obj_ptr_ != crh.cls_obj_ptr_;
}

void WMIObjectSet::next()
{
	if (enumerator_ptr_ != NULL) {
		unsigned long ret = 0;
		IWbemClassObject *ptr = NULL;
		HRESULT hr = enumerator_ptr_->Next(WBEM_INFINITE, 1, &ptr, &ret);
		if (SUCCEEDED(hr) && ret != 0) {
			cls_obj_ptr_ = ptr;
		}
	}
}

/////////////////////////////////////////////////////////////////////////////
// WMIService
/////////////////////////////////////////////////////////////////////////////

WMIService::WMIService(wchar_t const *obj_name)
: loc(NULL),
  svc(NULL)
{
	// Obtain the initial locator to WMI
	HRESULT hres = CoCreateInstance(
		CLSID_WbemLocator,
		0,
		CLSCTX_INPROC_SERVER,
		IID_IWbemLocator,
		reinterpret_cast<void**>(&loc));
	if (SUCCEEDED(hres)) {
		// Connect to WMI through the IWbemLocator::ConnectServer method
		//CComPtr<IWbemServices> svc = NULL;
		// Connect to the root\cimv2 namespace with
		// the current user and obtain pointer pSvc
		// to make IWbemServices calls.
		hres = loc->ConnectServer(
			_bstr_t(obj_name), // Object path of WMI namespace
			NULL,                    // User name. NULL = current user
			NULL,                    // User password. NULL = current
			NULL,                    // Locale. NULL indicates current
			0,                       // Security flags.
			NULL,                    // Authority (e.g. Kerberos)
			NULL,                    // Context object 
			&svc);                   // pointer to IWbemServices proxy
		if (SUCCEEDED(hres)) {
			// Set security levels on the proxy
			hres = CoSetProxyBlanket(
				svc,                        // Indicates the proxy to set
				RPC_C_AUTHN_WINNT,           // RPC_C_AUTHN_xxx
				RPC_C_AUTHZ_NONE,            // RPC_C_AUTHZ_xxx
				NULL,                        // Server principal name
				RPC_C_AUTHN_LEVEL_CALL,      // RPC_C_AUTHN_LEVEL_xxx
				RPC_C_IMP_LEVEL_IMPERSONATE, // RPC_C_IMP_LEVEL_xxx
				NULL,                        // client identity
				EOAC_NONE);                  // proxy capabilities
			//if (SUCCEEDED(hres)) {
			//}
		}
	}
}

WMIService::~WMIService()
{
	if (loc != NULL) {
		loc->Release();
	}
	if (svc != NULL) {
		svc->Release();
	}
}

IEnumWbemClassObject * WMIService::execQuery(wchar_t const *wql)
{
	// Use the IWbemServices pointer to make requests of WMI
	// For example, get the name of the operating system
	IEnumWbemClassObject *enumerator_ptr = NULL;
	/*HRESULT hres =*/ svc->ExecQuery(
		bstr_t("WQL"),
		bstr_t(wql),
		WBEM_FLAG_FORWARD_ONLY | WBEM_FLAG_RETURN_IMMEDIATELY,
		NULL,
		&enumerator_ptr);
	return enumerator_ptr;
}
