#pragma once

class BIOSInformation
{
public:
	std::wstring bios_vendor;
	std::wstring bios_caption;
	std::wstring bios_ver;
	std::wstring smbios_caption;
	unsigned int smbios_major_ver;
	unsigned int smbios_minor_ver;
	std::wstring mb_vendor;
	std::wstring mb_product;
	std::wstring mb_ver;
	std::wstring com_sys_vendor;
	std::wstring com_sys_model;

	BIOSInformation()
	{
		// Obtain the initial locator to WMI
		CComPtr<IWbemLocator> loc = NULL;
		HRESULT hres = loc.CoCreateInstance(
			CLSID_WbemLocator,
			0,
			CLSCTX_INPROC_SERVER);
		if (SUCCEEDED(hres)) {
			// Connect to WMI through the IWbemLocator::ConnectServer method
			CComPtr<IWbemServices> svc = NULL;
			// Connect to the root\cimv2 namespace with
			// the current user and obtain pointer pSvc
			// to make IWbemServices calls.
			hres = loc->ConnectServer(
				_bstr_t(L"ROOT\\CIMV2"), // Object path of WMI namespace
				NULL,                    // User name. NULL = current user
				NULL,                    // User password. NULL = current
				NULL,                    // Locale. NULL indicates current
				0,                       // Security flags.
				NULL,                    // Authority (e.g. Kerberos)
				NULL,                    // Context object 
				&svc);                   // pointer to IWbemServices proxy
			if (SUCCEEDED(hres)) {
				// Set security levels on the proxy
				hres = CoSetProxyBlanket(
					svc,                        // Indicates the proxy to set
					RPC_C_AUTHN_WINNT,           // RPC_C_AUTHN_xxx
					RPC_C_AUTHZ_NONE,            // RPC_C_AUTHZ_xxx
					NULL,                        // Server principal name
					RPC_C_AUTHN_LEVEL_CALL,      // RPC_C_AUTHN_LEVEL_xxx
					RPC_C_IMP_LEVEL_IMPERSONATE, // RPC_C_IMP_LEVEL_xxx
					NULL,                        // client identity
					EOAC_NONE);                  // proxy capabilities
				if (SUCCEEDED(hres)) {
					{
						// Use the IWbemServices pointer to make requests of WMI
						// For example, get the name of the operating system
						CComPtr<IEnumWbemClassObject> enum_obj = NULL;
						hres = svc->ExecQuery(
							bstr_t("WQL"),
							bstr_t("SELECT * FROM Win32_BIOS"),
							WBEM_FLAG_FORWARD_ONLY | WBEM_FLAG_RETURN_IMMEDIATELY,
							NULL,
							&enum_obj);
						unsigned long ret = 0;
						while (enum_obj) {
							CComPtr<IWbemClassObject> cls_obj;
							enum_obj->Next(WBEM_INFINITE, 1, &cls_obj, &ret);
							if (ret == 0) {
								break;
							}
							// Get the value of the Name property
							VARIANT v;
							{
								cls_obj->Get(L"Manufacturer", 0, &v, NULL, NULL); //
								this->bios_vendor = getStringValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"Caption", 0, &v, NULL, NULL); //
								this->bios_caption = getStringValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"Version", 0, &v, NULL, NULL); //
								this->bios_ver = getStringValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"SMBIOSBIOSVersion", 0, &v, NULL, NULL); //
								this->smbios_caption = getStringValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"SMBIOSMajorVersion", 0, &v, NULL, NULL); //
								this->smbios_major_ver = getIntValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"SMBIOSMinorVersion", 0, &v, NULL, NULL); //
								this->smbios_minor_ver = getIntValue(v);
								VariantClear(&v);
							}
						}
					}
					{
						// Use the IWbemServices pointer to make requests of WMI
						// For example, get the name of the operating system
						CComPtr<IEnumWbemClassObject> enum_obj = NULL;
						hres = svc->ExecQuery(
							bstr_t("WQL"),
							bstr_t("SELECT * FROM Win32_BaseBoard"),
							WBEM_FLAG_FORWARD_ONLY | WBEM_FLAG_RETURN_IMMEDIATELY,
							NULL,
							&enum_obj);
						unsigned long ret = 0;
						while (enum_obj) {
							CComPtr<IWbemClassObject> cls_obj;
							enum_obj->Next(WBEM_INFINITE, 1, &cls_obj, &ret);
							if (ret == 0) {
								break;
							}
							// Get the value of the Name property
							VARIANT v;
							{
								cls_obj->Get(L"Manufacturer", 0, &v, NULL, NULL); //
								this->mb_vendor = getStringValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"Product", 0, &v, NULL, NULL); //
								this->mb_product = getStringValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"Version", 0, &v, NULL, NULL); //
								this->mb_ver = getStringValue(v);
								VariantClear(&v);
							}
						}
					}
					{
						// Use the IWbemServices pointer to make requests of WMI
						// For example, get the name of the operating system
						CComPtr<IEnumWbemClassObject> enum_obj = NULL;
						hres = svc->ExecQuery(
							bstr_t("WQL"),
							bstr_t("SELECT * FROM Win32_ComputerSystem"),
							WBEM_FLAG_FORWARD_ONLY | WBEM_FLAG_RETURN_IMMEDIATELY,
							NULL,
							&enum_obj);
						unsigned long ret = 0;
						while (enum_obj) {
							CComPtr<IWbemClassObject> cls_obj;
							enum_obj->Next(WBEM_INFINITE, 1, &cls_obj, &ret);
							if (ret == 0) {
								break;
							}
							// Get the value of the Name property
							VARIANT v;
							{
								cls_obj->Get(L"Manufacturer", 0, &v, NULL, NULL); //
								this->com_sys_vendor = getStringValue(v);
								VariantClear(&v);
							}
							{
								cls_obj->Get(L"Model", 0, &v, NULL, NULL); //
								this->com_sys_model = getStringValue(v);
								VariantClear(&v);
							}
						}
					}
				}
			}
		}
	}

	std::wstring getStringValue(VARIANT const &v)
	{
		if (v.vt != VT_BSTR) {
			return L"";
		}
		return V_BSTR(&v);
	}

	unsigned int getIntValue(VARIANT const &v)
	{
		if (v.vt == VT_I2) {
			return  V_I2(&v);
		} else if (v.vt == VT_I4) {
			return V_I4(&v);
		} else if (v.vt == VT_UI1) {
			return V_UI1(&v);
		} else if (v.vt == VT_UI2) {
			return V_UI2(&v);
		} else if (v.vt == VT_UI4) {
			return V_UI4(&v);
		//} else if (v.vt == VT_I8) {
		//	return V_I8(&v);
		//} else if (v.vt == VT_UI8) {
		//	return V_UI8(&v);
		} else if (v.vt == VT_INT) {
			return V_INT(&v);
		} else if (v.vt == VT_UINT) {
			return V_UINT(&v);
		}
		return 0;
	}
};
