#ifndef OTHER_PROCESS_EXECUTOR_H__220CC9FD_19CF_4B0D_A042_069095B5CE23__INCLUDED
#define OTHER_PROCESS_EXECUTOR_H__220CC9FD_19CF_4B0D_A042_069095B5CE23__INCLUDED
// M$ compatible compilers support #pragma once
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

/////////////////////////////////////////////////////////////////////////////
// ProcessExecutor
//	vZX̋N
//  H23(2011)/11/12 쐬
/////////////////////////////////////////////////////////////////////////////

#include "Exception.h"

namespace MyLib {
class ProcessExecutor
{
public:
	// ~Ԃ̃AvP[V𓮍삳
	virtual void run(std::wstring const &command_line)
	{
		// vZXN
		// ł͒~ԂɂArun()œ삳
		size_t const buf_count = command_line.length() + 1;
		boost::scoped_array<wchar_t> buf(new wchar_t[buf_count]);
		memcpy(buf.get(), command_line.c_str(), buf_count*sizeof(wchar_t));
		PROCESS_INFORMATION process_info = {0};
		STARTUPINFO  startup_info = {sizeof (STARTUPINFO)};
		if (::CreateProcess(NULL, buf.get(), NULL,
			NULL, FALSE, 0, NULL, NULL, &startup_info, &process_info)) {
				this->process_handle.reset(process_info.hProcess);
				this->thread_handle.reset(process_info.hThread);
		} else {
			BOOST_THROW_EXCEPTION(Win32Exception());
			//throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}
	}

	unsigned int getStatus(unsigned int waiting_time_in_ms = 0)
	{
		return ::WaitForSingleObject(this->process_handle.get(), waiting_time_in_ms);
	}

	unsigned int getExitCode()
	{
		DWORD exit_code;
		if (!::GetExitCodeProcess(this->process_handle.get(), &exit_code)) {
			BOOST_THROW_EXCEPTION(Win32Exception());
			//throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}
		return exit_code;
	}

protected:
	AutoHandleN process_handle;
	AutoHandleN thread_handle;
};

class ConsoleProcessExecutor :
	public ProcessExecutor
{
public:
	// AvP[V𓮍삳
	virtual void run(std::wstring const &command_line)
	{
		// SECURITY_ATTRIBUTES ̐ݒ(pCv̂ɕKv)
		SECURITY_ATTRIBUTES security_attribs = {sizeof (SECURITY_ATTRIBUTES)};
		security_attribs.bInheritHandle = TRUE;		// nhp

		//pCv쐬
		HANDLE read_handle, write_handle;
		// W
		AutoHandleN stdin_read_pipe;
		if (::CreatePipe(&read_handle, &write_handle, &security_attribs, 0)) {
			stdin_read_pipe.reset(read_handle);
			AutoHandleN tmp_handle(write_handle);
			this->stdin_write_pipe.reset(getChildPipeHandle(write_handle));
		} else {
			BOOST_THROW_EXCEPTION(Win32Exception());
			//throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}
		// Wo
		AutoHandleN stdout_write_pipe;
		if (::CreatePipe(&read_handle, &write_handle, &security_attribs, 0)) {
			stdout_write_pipe.reset(write_handle);
			AutoHandleN tmp_handle(read_handle);
			this->stdout_read_pipe.reset(getChildPipeHandle(read_handle));
		} else {
			BOOST_THROW_EXCEPTION(Win32Exception());
			//throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}
		// WG[o
		AutoHandleN stderr_write_pipe;
		if (::CreatePipe(&read_handle, &write_handle, &security_attribs, 0)) {
			stderr_write_pipe.reset(write_handle);
			AutoHandleN tmp_handle(read_handle);
			this->stderr_read_pipe.reset(getChildPipeHandle(read_handle));
		} else {
			BOOST_THROW_EXCEPTION(Win32Exception());
			//throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}

		// vZXN
		// ł͒~ԂɂArun()œ삳
		size_t const buf_count = command_line.length() + 1;
		boost::scoped_array<wchar_t> buf(new wchar_t[buf_count]);
		memcpy(buf.get(), command_line.c_str(), buf_count*sizeof(wchar_t));
		PROCESS_INFORMATION process_info = {0};
		STARTUPINFO  startup_info = {sizeof (STARTUPINFO)};
		startup_info.dwFlags = STARTF_USESTDHANDLES;
		startup_info.hStdInput = stdin_read_pipe.get();
		startup_info.hStdOutput = stdout_write_pipe.get();
		startup_info.hStdError = stderr_write_pipe.get();
		if (::CreateProcess(NULL, buf.get(), &security_attribs,
			&security_attribs, TRUE, DETACHED_PROCESS, NULL, NULL,
			&startup_info, &process_info)) {
				this->process_handle.reset(process_info.hProcess);
				this->thread_handle.reset(process_info.hThread);
		} else {
			BOOST_THROW_EXCEPTION(Win32Exception());
			//throw MY_WIN32_EXCEPTION_LAST_ERROR;
		}
	}

	std::vector<char> readStdOut()
	{
		return read(this->stdout_read_pipe.get());
	}

	std::vector<char> readStdErr()
	{
		return read(this->stderr_read_pipe.get());
	}

	//HRESULT writeString(LPCTSTR str)
	//{
	//	ATLASSUME(this->stdinWritePipe.m_h);

	//	HRESULT result = S_OK;
	//	BOOL ret;
	//	CStringA sjis_str(str);
	//	DWORD written_bytes;
	//	ret = ::WriteFile(this->stdinWritePipe, sjis_str, sjis_str.GetLength(),
	//		&written_bytes, NULL);
	//	if (ret) {
	//		// CStringɉsǉƉs邱Ƃ̂ŁAsPƂœ
	//		ret = ::WriteFile(this->stdinWritePipe, "\r\n", 2, &written_bytes, NULL);
	//	}
	//	if (!ret) {
	//		result = HRESULT_FROM_WIN32(::GetLastError());
	//	}
	//	return result;
	//}

protected:
	// pCv̍쐬Ɏgp
	static HANDLE getChildPipeHandle(HANDLE h)
	{
		HANDLE result;
		bool ret = (::DuplicateHandle(::GetCurrentProcess(), h,
			::GetCurrentProcess(), &result, 0, FALSE,
			DUPLICATE_SAME_ACCESS) != 0);
		if (!ret) {
			BOOST_THROW_EXCEPTION(Win32Exception());
		}
		return result;
	}

	static std::vector<char> read(HANDLE named_pipe_handle)
	{
		std::vector<char> tmp_buf;
		DWORD avail_bytes;
		BOOL ret1 = ::PeekNamedPipe(named_pipe_handle, NULL, 0,
			NULL, &avail_bytes, NULL);
		if (ret1) {
			if (avail_bytes > 0) {
				tmp_buf.resize(avail_bytes);
				DWORD read_bytes;
				::ReadFile(named_pipe_handle,
					&tmp_buf[0], avail_bytes, &read_bytes, NULL);
			}
		//} else {
		//	DWORD err = ::GetLastError();
		//	BOOST_THROW_EXCEPTION(Win32Exception(err));
		}
		return tmp_buf;
	}

	AutoHandleN stdin_write_pipe;
	AutoHandleN stdout_read_pipe;
	AutoHandleN stderr_read_pipe;
};
};

#endif	// OTHER_PROCESS_EXECUTOR_H__220CC9FD_19CF_4B0D_A042_069095B5CE23__INCLUDED
