#pragma once

#include "Exception.h"

namespace MyLib {
//inline size_t convertUTF16ToUTF8(const wchar_t *s, char *d)
//inline std::string convertUTF16ToUTF8(const std::wstring& utf16_str)
//inline size_t convertUTF8ToUTF16(const char *s, wchar_t *d)
//inline std::wstring convertUTF8ToUTF16(const std::string& utf8_str)
//inline int convertToUTF8(LPCTSTR src, LPSTR dest, size_t dest_size)
//inline std::string convertToUTF8(const std::wstring& str)
//inline int convertToSJIS(LPCTSTR src, LPSTR dest, size_t dest_size)
//inline std::string convertToSJIS(const std::wstring& str)
//inline std::wstring convertFromUTF8(const std::string& str)
//inline std::wstring convertFromSJIS(const std::string& str)

inline size_t convertUTF16ToUTF8(const wchar_t *s, char *d)
{
	bool in_surrogate_pair = false;
	wchar_t high_surrogate = 0;
	size_t count = 1;
	while (*s != L'\0') {
		if (in_surrogate_pair) {
			if (IS_LOW_SURROGATE(*s)) {
				if (d != NULL) {
					uint32_t value =
						(((high_surrogate & 0x03FF) << 10) | (*s & 0x03FF))
						+ 0x010000;
					*d++ = static_cast<char>((value >> 18) | 0xF0);
					*d++ = static_cast<char>(((value >> 12) & 0x3F) | 0x80);
					*d++ = static_cast<char>(((value >> 6) & 0x3F) | 0x80);
					*d++ = static_cast<char>((value & 0x3F) | 0x80);
				}
				count += 4;
			} else {
				break;
			}
			in_surrogate_pair = false;
		} else {
			if (0x0000 <= *s && *s <= 0x007F) {
				if (d != NULL) {
					*d++ = static_cast<char>(*s);
				}
				count += 1;
			} else if (0x0080 <= *s && *s <= 0x07FF) {
				if (d != NULL) {
					*d++ = static_cast<char>(((*s >> 6) & 0x1F) | 0xC0);
					*d++ = static_cast<char>((*s & 0x3F) | 0x80);
				}
				count += 2;
			} else if (IS_HIGH_SURROGATE(*s)) {
				in_surrogate_pair = true;
				high_surrogate = *s;
			} else if (IS_LOW_SURROGATE(*s)) {
				break;
			} else if (0x0800 <= *s && *s <= 0xFFFF) {
				if (d != NULL) {
					*d++ = static_cast<char>((*s >> 12) | 0xE0);
					*d++ = static_cast<char>(((*s >> 6) & 0x3F) | 0x80);
					*d++ = static_cast<char>((*s & 0x3F) | 0x80);
				}
				count += 3;
			}
		}
		s++;
	}
	if (d != NULL) {
		*d = '\0';
	}
	return count;
}

inline std::string convertUTF16ToUTF8(const std::wstring& utf16_str)
{
	size_t n = convertUTF16ToUTF8(utf16_str.c_str(), NULL);
	char *ptr = new char[n];
	convertUTF16ToUTF8(utf16_str.c_str(), ptr);
	std::string result = ptr;
	delete[] ptr;
	return result;
}

inline size_t convertUTF8ToUTF16(const char *s, wchar_t *d)
{
	size_t count = 1;
	int flag = 0;
	uint32_t tmp = 0;
	while (*s != '\0') {
		unsigned char c = *s;
		if (flag == 0) {
			if ((c & 0xF8) == 0xF0) {	// 5byte11110
				flag = 3;
				tmp = static_cast<uint32_t>((c & 0x07)) << 18;
			} else if ((c & 0xF0) == 0xE0) {	// 4byte1110
				flag = 2;
				tmp = static_cast<uint32_t>((c & 0x0F)) << 12;
			} else if ((c & 0xE0) == 0xC0) {	// 3byte110
				flag = 1;
				tmp = static_cast<uint32_t>((c & 0x1F)) << 6;
			} else if ((c & 0x80) == 0) {	// 1byte0
				if (d != NULL) {
					*d++ = c;
				}
				++count;
			} else {
				break;
			}
		} else if (flag == 1) {
			flag = 0;
			if (tmp <= 0xFFFF) {
				if (d != NULL) {
					*d++ = static_cast<wchar_t>(tmp | (c & 0x3F));
				}
				++count;
			} else {
				if (d != NULL) {
					tmp |= (c & 0x3F);
					tmp -= 0x00010000;
					*d++ = static_cast<wchar_t>(((tmp >> 10) & 0x03FF) | HIGH_SURROGATE_START);
					*d++ = static_cast<wchar_t>((tmp & 0x03FF) | LOW_SURROGATE_START);
				}
				count += 2;
			}
		} else {
			--flag;
			tmp |= static_cast<uint32_t>((c & 0x3F)) << (flag*6);
		}
		s++;
	}
	if (d != NULL) {
		*d = L'\0';
	}
	return count;
}

inline std::wstring convertUTF8ToUTF16(const std::string& utf8_str)
{
	size_t n = convertUTF8ToUTF16(utf8_str.c_str(), NULL);
	wchar_t *ptr = new wchar_t[n];
	convertUTF8ToUTF16(utf8_str.c_str(), ptr);
	std::wstring result = ptr;
	delete[] ptr;
	return result;
}

//=======================================================================
// UTF-8ɕϊ(zev[g)
//=======================================================================
inline int convertToUTF8(LPCTSTR src, LPSTR dest, size_t dest_size)
{
#ifdef UNICODE
	int result_code = 0;
	const int ret = ::WideCharToMultiByte(CP_UTF8, 0, src, -1, dest,
		static_cast<int>(dest_size), NULL, NULL);
	if (ret == 0) {
		switch (::GetLastError()) {
			case ERROR_INSUFFICIENT_BUFFER:
				result_code = ERANGE;
				break;
			//case ERROR_INVALID_PARAMETER:
			//case ERROR_INVALID_FLAGS:
			default:
				result_code = EINVAL;
				break;
		}
	}
#else
#error Not implemented
#endif
	return result_code;
}
//=======================================================================
// UTF-8ɕϊ(std::string)
//=======================================================================
inline std::string convertToUTF8(const std::wstring& str)
{
#ifdef UNICODE
	std::string result;
	const int required_size = ::WideCharToMultiByte(CP_UTF8, 0,
		str.c_str(), -1, NULL, 0, NULL, NULL);
	if (required_size > 0) {
		std::vector<char> buf(required_size);
		const int ret = ::WideCharToMultiByte(CP_UTF8, 0,
			str.c_str(), -1, &buf[0], required_size, NULL, NULL);
		if (ret != 0) {
			result = &buf[0];
		} else {
			BOOST_THROW_EXCEPTION(Win32Exception());
		}
	} else {
		BOOST_THROW_EXCEPTION(Win32Exception());
	}
	return result;
#else
#error Not implemented
#endif
}

//=======================================================================
// ShiftJIS(CP932)ɕϊ(zev[g)
//=======================================================================
inline int convertToSJIS(LPCTSTR src, LPSTR dest, size_t dest_size)
{
#ifdef UNICODE
	int result_code = 0;
	BOOL is_default_char_used;
	const int ret = ::WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS,
		src, -1, dest, static_cast<int>(dest_size), NULL, &is_default_char_used);
	if (ret == 0) {
		switch (::GetLastError()) {
			case ERROR_INSUFFICIENT_BUFFER:
				result_code = ERANGE;
				break;
			//case ERROR_INVALID_PARAMETER:
			//case ERROR_INVALID_FLAGS:
			default:
				result_code = EINVAL;
				break;
		}
	} else if (is_default_char_used) {
		result_code = EILSEQ;
	}
#else
#error Not implemented
#endif
	return result_code;
}
//=======================================================================
// ShiftJIS(CP932)ɕϊ(std::string)
//=======================================================================
inline std::string convertToSJIS(const std::wstring& str)
{
#ifdef UNICODE
	std::string result;
	const int required_size = ::WideCharToMultiByte(CP_ACP,
		WC_NO_BEST_FIT_CHARS, str.c_str(), -1, NULL, 0, NULL, NULL);
	if (required_size > 0) {
		std::vector<char> buf(required_size);
		BOOL is_default_char_used;
		const int ret = ::WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS,
			str.c_str(), -1, &buf[0], required_size, NULL, &is_default_char_used);
		if (ret == 0) {
			BOOST_THROW_EXCEPTION(Win32Exception());
		} else if (is_default_char_used) {
			BOOST_THROW_EXCEPTION(StdCException(EILSEQ));
		}
		result = &buf[0];
	} else {
		BOOST_THROW_EXCEPTION(Win32Exception());
	}
	return result;
#else
#error Not implemented
#endif
}

inline std::wstring convertFromUTF8(const std::string& str)
{
	std::wstring result;
#ifdef UNICODE
	const int required_size = ::MultiByteToWideChar(CP_UTF8, 0, str.c_str(),
		-1, NULL, 0);
	if (required_size > 0) {
		std::vector<TCHAR> buf(required_size);
		const int ret = ::MultiByteToWideChar(
			CP_UTF8, 0, str.c_str(), -1, &buf[0], required_size);
		if (ret == 0) {
			BOOST_THROW_EXCEPTION(Win32Exception());
		}
		result = &buf[0];
	}
#else
#error Not implemented
#endif
	return result;
}

inline std::wstring convertFromSJIS(const std::string& str)
{
	std::wstring result;
#ifdef UNICODE
	const int required_size = ::MultiByteToWideChar(
		CP_ACP, MB_ERR_INVALID_CHARS, str.c_str(), -1, NULL, 0);
	if (required_size > 0) {
		std::vector<TCHAR> buf(required_size);
		const int ret = ::MultiByteToWideChar(
			CP_ACP, 0, str.c_str(), -1, &buf[0], required_size);
		if (ret == 0) {
			BOOST_THROW_EXCEPTION(Win32Exception());
		}
		result = &buf[0];
	}
#else
#error Not implemented
#endif
	return result;
}
}; // namespace MyLib
